/*
Copyright (c) 2006-2007, Group 1A TI ProHES 2006/2007
                         Fontys University of Professional Education
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/* Borland stuff, delete if you don't use this product from Hell */
#pragma hdrstop
#pragma package(smart_init)

/* Standard headers */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "bliepdsp.h"

unsigned int bliepdsp::calculate_time (void) {
  return this->header.data_size / (this->header.num_channels * this->header.sample_rate * (this->header.bits_per_sample / 8));
}

void bliepdsp::read_wav (const char *filename) {
  FILE *rawwav = fopen(filename, "rb");

  if (rawwav == NULL) {
    fprintf(stderr, "Could not open PCM WAV file: %s\n", filename);
    this->data = NULL;
    return;
  }
  fread(&this->header, sizeof(struct wavheader), 1, rawwav);
  if (this->header.id == PCM_RIFF && this->header.format == PCM_FORMAT && this->header.audio_format == 1) {
    this->data = (char *) malloc(this->header.data_size);
    if (this->data == NULL) {
      fprintf(stderr, "Out of memory: %s\n", filename);
      return;
    }
    fseek(rawwav, sizeof(struct wavheader), SEEK_SET);
    fread(this->data, 1, this->header.data_size, rawwav);
  } else
    this->data = NULL;

  this->num_samples = this->header.data_size / (this->header.bits_per_sample / 8) / this->header.num_channels;
  fclose(rawwav);
}

void bliepdsp::freedata (void) {
  if (this->data != NULL)
    free(this->data);
}

void bliepdsp::write_wav (const char *filename) {
  FILE *outwav = fopen(filename, "wb+");
  if (outwav == NULL) {
    fprintf(stderr, "Could not write to file: %s\n", filename);
    return;
  }
  fwrite(&this->header, 1, sizeof(struct wavheader), outwav);
  fwrite(this->data, 1, this->header.data_size, outwav);
  fclose(outwav);
}

void bliepdsp::filter_echo (bliepdsp *out, const unsigned int delay, const unsigned int wet) {
  if (this->header.bits_per_sample < 8)
    return;

  out->header = this->header;
  unsigned int bytes_per_sample = out->header.bits_per_sample / 8;
  unsigned int delayed_samples = (out->header.sample_rate / 1000) * delay; /* not exact */
  double lower = (double) wet / 100;
  double upper = (double) (100 - wet) / 100;
  out->header.data_size = this->header.data_size + out->header.num_channels * bytes_per_sample * delayed_samples;
  out->data = (char *) malloc(out->header.data_size);
  if (out->data == NULL)
    return;

  memcpy(out->data, this->data, this->header.data_size);

  unsigned int i;
  short int buf1_16, buf2_16, buf3_16, buf3_8;
  unsigned char buf1_8, buf2_8, buf4_8;
  char *source, *target;
  for (i = bytes_per_sample * delayed_samples * out->header.num_channels; i < out->header.data_size; i += bytes_per_sample) {
    source = (char *) this->data + (i - bytes_per_sample * delayed_samples * out->header.num_channels);
    target = (char *) out->data + i;
    switch (out->header.bits_per_sample) {
      case 8:
        memcpy(&buf1_8, source, bytes_per_sample);
        memcpy(&buf2_8, target, bytes_per_sample);
        buf3_8 = (buf1_8 - 128) * lower + (buf2_8 - 128) * upper;
        buf4_8 = buf3_8 + 128;
        memcpy(target, &buf4_8, bytes_per_sample);
      break;
      case 16:
        memcpy(&buf1_16, source, bytes_per_sample);
        memcpy(&buf2_16, target, bytes_per_sample);
        buf3_16 = buf1_16 * lower + buf2_16 * upper;
        memcpy(target, &buf3_16, bytes_per_sample);
      break;
      /* 24 bit (3B) and 32 bit (float) not yet supported */
      case 24:
      break;
      case 32:
      break;
    }
  }
}

void bliepdsp::filter_reverb (bliepdsp *out, const unsigned int delay, const unsigned int wet, const unsigned int minimp) {
  if (this->header.bits_per_sample < 8)
    return;

  out->header = this->header;
  unsigned int bytes_per_sample = out->header.bits_per_sample / 8;
  unsigned int delayed_samples = (out->header.sample_rate / 1000) * delay; /* not exact */
  double lower = (double) wet / 100;
  out->header.data_size = this->header.data_size + out->header.num_channels * bytes_per_sample * delayed_samples;
  out->data = (char *) malloc(out->header.data_size);
  if (out->data == NULL)
    return;

  memcpy(out->data, this->data, this->header.data_size);

  unsigned int i;
  short int buf1_16, buf3_16, buf3_8;
  unsigned short int trap_value16, trap_check16;
  unsigned char buf1_8, buf4_8, trap_value8, trap_check8, cutoff = 1;
  char *source, *target;
  for (i = this->header.data_size; i < out->header.data_size; i += bytes_per_sample) {
    source = (char *) out->data + i - out->header.num_channels * bytes_per_sample * delayed_samples;
    target = (char *) out->data + i;
    switch (out->header.bits_per_sample) {
      case 8:
        memcpy(&buf1_8, source, bytes_per_sample);
        buf3_8 = (buf1_8 - 128) * lower;
        buf4_8 = buf3_8 + 128;
        memcpy(target, &buf4_8, bytes_per_sample);
        trap_value8 = buf4_8;
        trap_check8 = 0xFF * ((double) minimp / 100);
        if (trap_value8 > trap_check8)
          cutoff = 0;
      break;
      case 16:
        memcpy(&buf1_16, source, bytes_per_sample);
        buf3_16 = buf1_16 * lower;
        memcpy(target, &buf3_16, bytes_per_sample);
        trap_value16 = fabs(buf3_16);
        trap_check16 = 0x7FFF * ((double) minimp / 100);
        if (trap_value16 > trap_check16)
          cutoff = 0;
      break;
      /* 24 bit (3B) and 32 bit (float) not yet supported */
      case 24:
      break;
      case 32:
      break;
    }
    if (i + bytes_per_sample == out->header.data_size && cutoff == 0) {
      cutoff = 1;
      out->header.data_size += out->header.num_channels * bytes_per_sample * delayed_samples;
      out->data = (char *) realloc(out->data, out->header.data_size);
      if (out->data == NULL)
        return;
    }
  }
}

void bliepdsp::filter_backwards (bliepdsp *out) {
  if (this->header.bits_per_sample < 8)
    return;

  out->header = this->header;
  unsigned int bytes_per_multichannel_sample = (out->header.bits_per_sample / 8) * out->header.num_channels;
  char *source, *target;

  out->data = (char *) malloc(out->header.data_size);
  if (out->data == NULL)
    return;

  unsigned int i;
  for (i = out->header.data_size - bytes_per_multichannel_sample; i > 0; i -= bytes_per_multichannel_sample) {
    source = (char *) data + this->header.data_size - bytes_per_multichannel_sample - i;
    target = (char *) out->data + i;
    memcpy(target, source, bytes_per_multichannel_sample);
  }
}

void bliepdsp::filter_stretch (bliepdsp *out, const unsigned int stretch) {
  if (this->header.bits_per_sample < 8)
    return;

  out->header = this->header;
  unsigned int bytes_per_multichannel_sample = (out->header.bits_per_sample / 8) * out->header.num_channels;
  char *source, *target;
  unsigned int num_samples_till_stretch, i, j = 1, stretch_samples = 0;

  if (stretch > 100) {
    num_samples_till_stretch = (unsigned int) floor(100 / (double) (stretch - 100));
    for (i = 0; i < this->header.data_size; i += bytes_per_multichannel_sample) {
      if (j == num_samples_till_stretch) {
        out->header.data_size += bytes_per_multichannel_sample;
        j = 1;
      } else
        j++;
    }

    out->data = (char *) malloc(out->header.data_size);
    if (out->data == NULL)
      return;

    j = 1;
    for (i = 0; i < this->header.data_size; i += bytes_per_multichannel_sample) {
      source = (char *) this->data + i;
      target = (char *) out->data + i + stretch_samples * bytes_per_multichannel_sample;
      if (j == num_samples_till_stretch) {
        memcpy(target, source, bytes_per_multichannel_sample);
        stretch_samples++;
        target = (char *) out->data + i + stretch_samples * bytes_per_multichannel_sample;
        j = 1;
      } else
        j++;
      memcpy(target, source, bytes_per_multichannel_sample);
    }
  } else if (stretch < 100) {
    num_samples_till_stretch = (unsigned int) floor(100 / stretch);
    for (i = 0; i < this->header.data_size; i += bytes_per_multichannel_sample) {
      if (j == num_samples_till_stretch) {
        out->header.data_size -= bytes_per_multichannel_sample;
        j = 1;
      } else
        j++;
    }

    out->data = (char *) malloc(out->header.data_size);
    if (out->data == NULL)
      return;

    for (i = 0; i < this->header.data_size; i += bytes_per_multichannel_sample) {
      if (j == num_samples_till_stretch) {
        stretch_samples++;
        j = 1;
      } else
        j++;
      source = (char *) this->data + i;
      target = (char *) out->data + i - (stretch_samples * bytes_per_multichannel_sample);
      memcpy(target, source, bytes_per_multichannel_sample);
    }
  }
  out->num_samples = out->header.data_size / (out->header.bits_per_sample / 8) / out->header.num_channels;
}

void bliepdsp::copy (bliepdsp *out, const unsigned int start, const unsigned int stop) {
  unsigned int i, bytes_per_multichannel_sample;
  char *source, *target;
  bytes_per_multichannel_sample = this->header.num_channels * (this->header.bits_per_sample / 8);
  out->header = this->header;
  out->header.data_size = (stop - start) * bytes_per_multichannel_sample;
  out->data = (char *) malloc(out->header.data_size);
  if (out->data == NULL)
    return;

  target = out->data - bytes_per_multichannel_sample;
  for (i = start; i <= stop; i++) {
    source = this->data + i * bytes_per_multichannel_sample;
    target += bytes_per_multichannel_sample;
    memcpy(target, source, bytes_per_multichannel_sample);
  }
  out->num_samples = out->header.data_size / (out->header.bits_per_sample / 8) / out->header.num_channels;
}

void bliepdsp::cut (bliepdsp *out, const unsigned int start, const unsigned int stop) {
  unsigned int i, bytes_per_multichannel_sample;
  char *source, *target, *buffer, *bufferc;
  bytes_per_multichannel_sample = this->header.num_channels * (this->header.bits_per_sample / 8);
  out->header = this->header;
  out->header.data_size = (stop - start) * bytes_per_multichannel_sample;
  out->data = (char *) malloc(out->header.data_size);
  if (out->header.data_size != this->header.data_size) {
    buffer = (char *) malloc(this->header.data_size - out->header.data_size);
    if (buffer == NULL)
      return;
  }
  if (out->data == NULL)
    return;

  bufferc = buffer;
  for (i = 0; i < this->header.data_size / bytes_per_multichannel_sample; i++) {
    source = this->data + i * bytes_per_multichannel_sample;
    if (i >= start && i <= stop)
      target = out->data + (i - start) * bytes_per_multichannel_sample;
    else {
      buffer += bytes_per_multichannel_sample;
      target = buffer;
    }
    memcpy(target, source, bytes_per_multichannel_sample);
  }
  this->header.data_size -= out->header.data_size;
  this->data = (char *) realloc(this->data, this->header.data_size);
  if (this->data == NULL)
    return;
  memcpy(this->data, bufferc, this->header.data_size);
  free(bufferc);
  out->num_samples = out->header.data_size / (out->header.bits_per_sample / 8) / out->header.num_channels;
}

void bliepdsp::paste (bliepdsp *in, const unsigned int where) {
  unsigned int i, bytes_per_multichannel_sample;
  char *source, *target, *buffer;
  bytes_per_multichannel_sample = this->header.num_channels * (this->header.bits_per_sample / 8);
  this->header.data_size += in->header.data_size;
  this->data = (char *) realloc(this->data, this->header.data_size);
  buffer = (char *) malloc(this->header.data_size - where * bytes_per_multichannel_sample);
  if (this->data == NULL || buffer == NULL)
    return;

  memcpy(buffer, this->data + where * bytes_per_multichannel_sample, this->header.data_size - where * bytes_per_multichannel_sample);

  for (i = where * bytes_per_multichannel_sample; i < this->header.data_size; i += bytes_per_multichannel_sample) {
    target = this->data + i;
    if (i <= where * bytes_per_multichannel_sample + in->header.data_size)
      source = in->data + i - where * bytes_per_multichannel_sample;
    else
      source = buffer + i - where * bytes_per_multichannel_sample - in->header.data_size;

    memcpy(target, source, bytes_per_multichannel_sample);
  }
  this->num_samples = this->header.data_size / (this->header.bits_per_sample / 8) / this->header.num_channels;
}

void bliepdsp::fade_in (bliepdsp *out, const unsigned int fadetime) {
  short int buf16;
  unsigned char buf8;
  char *source, *target;
  int end_fade, j;

  if (this->header.bits_per_sample < 8)
    return;

  out->header = this->header;
  out->data = (char *) malloc(out->header.data_size);
  if (out->data == NULL)
    return;

  memcpy(out->data, this->data, out->header.data_size);
  end_fade = (out->header.sample_rate * ((double) fadetime / 1000)) * out->header.num_channels * (out->header.bits_per_sample / 8);

  for (int i = 0; i < end_fade; i += (this->header.bits_per_sample / 8)) {
    source = this->data + i;
    target = out->data + i;

    switch (out->header.bits_per_sample) {
      case 8:
        memcpy(&buf8, source, 1);
        buf8 = ((buf8 - 128) * (double) i / end_fade) + 128;
        memcpy(target, &buf8, 1);
      break;
      case 16:
        memcpy(&buf16, source, 2);
        buf16 = buf16 * ((double) i / end_fade);
        memcpy(target, &buf16, 2);
      break;
      /* 24 bit (3B) and 32 bit (float) not yet supported  */
      case 24:
      break;
      case 32:
      break;
    }
  }
}

void bliepdsp::fade_out (bliepdsp *out, const unsigned int fadetime) {
  short int buf16;
  unsigned char buf8;
  char *source, *target;
  unsigned int start_fade;
  int j;

  if (this->header.bits_per_sample < 8)
    return;

  out->header = this->header;
  out->data = (char *) malloc(out->header.data_size);
  if (out->data == NULL)
    return;

  memcpy(out->data, this->data, out->header.data_size);
  start_fade = out->header.sample_rate * ((double) fadetime / 1000) * out->header.num_channels * (out->header.bits_per_sample / 8); /* FIXME: not exact */

  for (unsigned int i = out->header.data_size; i > start_fade; i -= (this->header.bits_per_sample / 8)) {
    source = this->data + i;
    target = out->data + i;

    switch (out->header.bits_per_sample) {
      case 8:
        memcpy(&buf8, source, 1);
        buf8 = ((buf8 - 128) *  (1 - ((double) i - start_fade) / (out->header.data_size - start_fade))) + 128;
        memcpy(target, &buf8, 1);
      break;
      case 16:
        memcpy(&buf16, source, 2);
        buf16 = buf16 * (1 - (((double) i - start_fade) / (out->header.data_size - start_fade)));
        memcpy(target, &buf16, 2);
      break;
      /* 24 bit (3B) and 32 bit (float) not yet supported  */
      case 24:
      break;
      case 32:
      break;
    }
  }
}

void bliepdsp::filter_recursive (bliepdsp *out, const unsigned int filter, const unsigned int co) {
  if (this->header.bits_per_sample != 8 && this->header.bits_per_sample != 16)
    return;
  out->header = this->header;
  out->data = (char *) malloc(out->header.data_size);
  memcpy(out->data, this->data, this->header.data_size);

  double a0, a1, a2, b1, b2, k, r, bandwidth, cosi, centerfreq;
  switch (filter) {
    case LOW_PASS:
      a0 = 1.00 - (double) co / 100;
      a1 = 0;
      a2 = 0;
      b1 = (double) co / 100;
      b2 = 0;
    break;
    case HIGH_PASS:
      a0 = (1.00 + (double) co / 100) / 2.00;
      a1 = -(1.00 + (double) co / 100) / 2.00;
      a2 = 0;
      b1 = (double) co / 100;
      b2 = 0;
    break;
    case NARROW_PASS:
      bandwidth = 0.0006;
      centerfreq = (double) co / (double) this->header.sample_rate;
      cosi = (double) cos(2 * PI * centerfreq);
      r = 1 - 3 * bandwidth;
      k = (1 - 2 * r * cosi + r * r) / (2 - 2 * cosi);
      a0 = 1 - k;
      a1 = 2 * (k - r) * cosi;
      a2 = r * r - k;
      b1 = 2 * r * cosi;
      b2 = -(r * r);
    break;
    case NARROW_REJECT:
      bandwidth = 0.0006;
      centerfreq = (double) co / (double) this->header.sample_rate;
      cosi = (double) cos(2 * PI * centerfreq);
      r = 1 - 3 * bandwidth;
      k = (1 - 2 * r * cosi + r * r) / (2 - 2 * cosi);
      a0 = k;
      a1 = -(2 * k * cosi);
      a2 = k;
      b1 = 2 * r * cosi;
      b2 = -(r * r);
    break;
  }

  unsigned int i, bytes_per_sample = this->header.bits_per_sample / 8;
  int c;
  unsigned char ca0, ca1, ca2, cb1, cb2, cout;
  short int ia0, ia1, ia2, ib1, ib2, ca0_s, ca1_s, ca2_s, cb1_s, cb2_s, iout, cout_s;
  char *target, *source;
  for (c = 0; c < this->header.num_channels; c++) { /* For every channel */
    for (i = c * bytes_per_sample + 2 * this->header.num_channels * bytes_per_sample;
         i < this->header.data_size; i += this->header.num_channels * bytes_per_sample) {
      /* For every sample within that channel, starting from the third do:
         out[i] = a0 * in[i] + a1 * in[i - 1] + a2 * in[i - 2] + b1 * out[i - 1] + b2 * out[i - 2] */
      target = (char *) out->data + i;
      source = (char *) this->data + i;
      switch (this->header.bits_per_sample) {
        case 8:
          memcpy(&ca0, source, bytes_per_sample);
          memcpy(&ca1, source - 1 * this->header.num_channels * bytes_per_sample, bytes_per_sample);
          memcpy(&ca2, source - 2 * this->header.num_channels * bytes_per_sample, bytes_per_sample);
          memcpy(&cb1, target - 1 * this->header.num_channels * bytes_per_sample, bytes_per_sample);
          memcpy(&cb2, target - 2 * this->header.num_channels * bytes_per_sample, bytes_per_sample);
          ca0_s = ca0 - 128;
          ca1_s = ca1 - 128;
          ca2_s = ca2 - 128;
          cb1_s = cb1 - 128;
          cb2_s = cb2 - 128;
          cout_s = a0 * ca0_s + a1 * ca1_s + a2 * ca2_s + b1 * cb1_s + b2 * cb2_s;
          cout = cout_s + 128;
          memcpy(target, &cout, bytes_per_sample);
        break;
        case 16:
          memcpy(&ia0, source, bytes_per_sample);
          memcpy(&ia1, source - 1 * this->header.num_channels * bytes_per_sample, bytes_per_sample);
          memcpy(&ia2, source - 2 * this->header.num_channels * bytes_per_sample, bytes_per_sample);
          memcpy(&ib1, target - 1 * this->header.num_channels * bytes_per_sample, bytes_per_sample);
          memcpy(&ib2, target - 2 * this->header.num_channels * bytes_per_sample, bytes_per_sample);
          iout = a0 * ia0 + a1 * ia1 + a2 * ia2 + b1 * ib1 + b2 * ib2;
          memcpy(target, &iout, bytes_per_sample);
        break;
        /* 24 bit (3B) and 32 bit (float) not yet supported */
        case 24:
        break;
        case 32:
        break;
      }
    }
  }
  out->num_samples = out->header.data_size / (out->header.bits_per_sample / 8) / out->header.num_channels;
}

