/*
Copyright (c) 2006-2007, Group 1A TI ProHES 2006/2007
                         Fontys University of Professional Education
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/* Borland stuff, delete if you don't use this product from Hell */
#pragma hdrstop
#pragma package(smart_init)

/* Standard headers */
#include "bliepdsp.h"
#include "bliepdsp_sdl.h"
#include <math.h>

/* SDL headers */
#include "SDL/SDL.h"
#include "SDL/SDL_main.h"
#include "SDL/SDL_mixer.h"

unsigned char bliep_init = 0;
unsigned int bliep_playsince = 0;
unsigned int bliep_pausesince = 0;
Mix_Chunk *sdl_wav;
struct {
  unsigned int num_samples;
  unsigned int length;
  unsigned short int bytes_per_sample;
  unsigned short int num_channels;
  char *data;
} bliep_sampleinfo;

int bliep_query_playing (void) {
  if (bliep_init)
    return Mix_Playing(1);
  return 0;
}

int bliep_query_paused (void) {
  if (bliep_init)
    return Mix_Paused(1);
  return 0;
}

void bliep_pause (void) {
  if (bliep_init) {
    Mix_Pause(1);
    bliep_pausesince = SDL_GetTicks();
  }
}

void bliep_unpause (void) {
  if (bliep_init) {
    Mix_Resume(1);
    bliep_playsince += (SDL_GetTicks() - bliep_pausesince);
  }
}

void bliep_stop (void) {
  if (bliep_init)
    Mix_HaltChannel(1);
}

void bliep_cleanup (int channel) {
  if (bliep_init) {
    Mix_FreeChunk(sdl_wav);
    bliep_init = 0;
  }
}

unsigned int bliep_currentsample (void) {
  if (!bliep_init)
    return 0;

  double progress = (double) bliep_currenttime() / bliep_sampleinfo.length;
  if (progress <= 1)
    return (unsigned int) floor(bliep_sampleinfo.num_samples * progress);
  else
    return bliep_sampleinfo.num_samples;
}

unsigned int bliep_currenttime (void) {
  if (!bliep_init)
    return 0;

  return SDL_GetTicks() - bliep_playsince;
}

bliep_sampledata bliep_getsampledata (void) {
  bliep_sampledata out;
  if (!bliep_init) {
    out.left = 0;
    out.right = 0;
    return out;
  }
  unsigned char ctmp;
  unsigned short int itmp;
  char *c = bliep_sampleinfo.data + bliep_currentsample() * bliep_sampleinfo.bytes_per_sample * bliep_sampleinfo.num_channels;
  if (bliep_sampleinfo.bytes_per_sample == 1) {
    memcpy(&ctmp, c, bliep_sampleinfo.bytes_per_sample);
    out.left = (unsigned short int) floor((double) fabs(ctmp - 128) / 128 * 1000);
  } else {
    memcpy(&itmp, c, bliep_sampleinfo.bytes_per_sample);
    out.left = (unsigned short int) floor((double) fabs(itmp) / 32768 * 1000);
  }


  if (bliep_sampleinfo.num_channels == 1) {
    out.right = out.left;
  } else {
    c += bliep_sampleinfo.bytes_per_sample;
    if (bliep_sampleinfo.bytes_per_sample == 1) {
      memcpy(&ctmp, c, bliep_sampleinfo.bytes_per_sample);
      out.right = (unsigned short int) floor((double) fabs(ctmp - 128) / 128 * 1000);
    } else {
      memcpy(&itmp, c, bliep_sampleinfo.bytes_per_sample);
      out.right = (unsigned short int) floor((double) fabs(itmp) / 32768 * 1000);
    }
  }

  return out;
}

void bliep_play (bliepdsp *wav) {
  if (bliep_init)
    return;

  sdl_wav = (Mix_Chunk *) malloc(sizeof(Mix_Chunk));
  bliep_sampleinfo.bytes_per_sample = wav->header.bits_per_sample / 8;
  bliep_sampleinfo.num_samples = wav->header.data_size / bliep_sampleinfo.bytes_per_sample / wav->header.num_channels;
  bliep_sampleinfo.length = wav->calculate_time() * 1000;
  bliep_sampleinfo.data = wav->data;
  bliep_sampleinfo.num_channels = wav->header.num_channels;

  unsigned int format = (wav->header.bits_per_sample == 16) ? AUDIO_S16SYS : AUDIO_U8;
  Mix_OpenAudio(wav->header.sample_rate, format, wav->header.num_channels, SDL_CHUNK_SIZE);
  bliep_init = 1;
  sdl_wav->allocated = 1;
  sdl_wav->abuf = wav->data;
  sdl_wav->alen = wav->header.data_size;
  sdl_wav->volume = 128;
  Mix_PlayChannel(1, sdl_wav, 0);
  bliep_playsince = SDL_GetTicks();
  Mix_ChannelFinished(bliep_cleanup);
}

bliepdsp bliep_decode_file (char *filename) {
  char default_header[44] = { /* PCM WAV 16bit 44100KHz 2channels */
    0x52, 0x49, 0x46, 0x46, 0x24, 0xF6, 0x97, 0x02,
    0x57, 0x41, 0x56, 0x45, 0x66, 0x6D, 0x74, 0x20,
    0x10, 0x00, 0x00, 0x00, 0x01, 0x00, 0x02, 0x00,
    0x44, 0xAC, 0x00, 0x00, 0x10, 0xB1, 0x02, 0x00,
    0x04, 0x00, 0x10, 0x00, 0x64, 0x61, 0x74, 0x61,
    0x00, 0xF6, 0x97, 0x02
  };
  bliepdsp out;
  Mix_Chunk *converted; /* Possible memory leak (a few bytes) when not freeing this one. Freeing it destroys the bliepdsp object. */
  Mix_OpenAudio(44100, AUDIO_S16SYS, 2, SDL_CHUNK_SIZE);
  converted = Mix_LoadWAV(filename);
  if (converted == NULL) {
    Mix_CloseAudio();
    out.data = NULL;
    return out;
  }
  out.data = converted->abuf;
  memcpy(&out.header, default_header, 44);
  out.header.data_size = converted->alen;
  out.num_samples = out.header.data_size / 4;
  Mix_CloseAudio();
  return out;
}
