/*
Copyright (c) 2006-2007, Group 1A TI ProHES 2006/2007
                         Fontys University of Professional Education
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/**
 * @file bliepdsp_sdl.h
 * @brief SDL layer for BliepDSP
 *
 * Uses SDL_Mixer (http://www.libsdl.org/projects/SDL_mixer/).
 *
 * ToDo:
 *  - Use SDL_Sound for codec support.
 *  - Try to find a more accurate way of determening current sample.
 *
 * @warning For SDL with Borland:
 *          Check at the compiler options "Treat enum types as ints".
 */

#ifndef BLIEPDSP_SDLH
#define BLIEPDSP_SDLH
/** @brief Fix SDL WIN32 Borland. */
#define _SIZE_T_DEFINED_
/** @brief Chunk buffer size. */
#define SDL_CHUNK_SIZE 2048

/**
 * @brief Current sample data in the playing audio sample.
 *
 * Volume in promille.
 */
struct bliep_sampledata {
  unsigned short int left; /**< Left audio channel. */
  unsigned short int right; /**< Right audio channel. */
};

/**
 * @brief Query if it is playing audio.
 *
 * 0 = stopped.
 * 1 = playing.
 */
int bliep_query_playing (void);

/**
 * @brief Query if it is paused.
 *
 * 0 = stopped or playing.
 * 1 = paused.
 */
int bliep_query_paused (void);

/**
 * @brief Pause audio.
 *
 * Pause current playing sample.
 * @see bliep_query_paused()
 */
void bliep_pause (void);

/**
 * @brief Unpause audio.
 *
 * Unpause current playing sample.
 * @see bliep_query_paused()
 */
void bliep_unpause (void);

/**
 * @brief Stop audio playback.
 *
 * Stop current playing sample.
 * @see bliep_query_playing()
 */
void bliep_stop (void);

/**
 * @brief Play audio.
 *
 * Play current selected sample.
 *
 * @param wav The bliepdsp object to play.
 *
 * @see bliep_query_playing()
 */
void bliep_play (bliepdsp *wav);

/**
 * @brief Return current sample number.
 *
 * Only if playing, else return 0.
 * @see bliep_query_playing()
 */
unsigned int bliep_currentsample (void);

/**
 * @brief Populates sampledata with current sample in audiosample.
 *
 * Even if it is paused, it will return the current sample in buffer.
 */
bliep_sampledata bliep_getsampledata (void);

/**
 * @brief Return current time in milliseconds.
 *
 * Only if playing, else return 0.
 * @see bliep_query_playing()
 */
unsigned int bliep_currenttime (void);

/**
 * @brief Convert codec'd file to internal PCM uncompressed WAV format
 *
 * Currently only supports the SDL_mixer supported codecs:
 *  - WAVE
 *  - AIFF
 *  - RIFF
 *  - OGG
 *  - VOC
 * Returns NULL on error.
 *
 * @warning This defaults to 16 bit stereo 44.1KHz.
 *          You might loose audio data.
 *
 * @param filename Path to encoded file.
 * @see bliepdsp::read_wav()
 */
bliepdsp bliep_decode_file (char *filename);

#endif
